import numpy as np
import numbers
import paired_data_helpers as pdh
import random_sampler
from scipy.stats import rankdata


def threshold(x, t):
        return np.sign(x) * np.clip(np.abs(x) - t, a_min=0, a_max=np.Inf)


def binary_solver(f, lower, upper, tol = 1e-10):
    """
    Returns x such that f(x) = 0
    :param f: Univariate, monotone decreasing function.
    :param lower: Lower bound on solution x.
    :param upper: Upper bound on solution x.
    :return: Numeric solving f(x) = 0
    """

    x = (lower + upper)/2
    val = f(x)
    if np.abs(val) < tol:
        return x
    elif val > 0:
        return binary_solver(f, x, upper)
    else:
        return binary_solver(f, lower, x)


def _clipped_l1_norm(x, c, lower, upper):
    return np.sum(np.clip(x-c, a_min=lower, a_max=upper))


def project_simplex_with_box_constraints(x, l1_norm=1, linf=np.Infinity):

    assert l1_norm > 0, "Norm constraint must be positive."
    assert linf > 0, "Norm constraint must be positive."

    if linf >= l1_norm: return project_positive_simplex(x, l1_norm)

    xclip = np.clip(x, a_min=0, a_max=linf)
    if np.sum(xclip) <= l1_norm:
        return xclip

    MM = np.max(x)

    mu = binary_solver(lambda u: _clipped_l1_norm(x, u, 0, linf) - l1_norm, 0, MM)
    return np.clip(x - mu, a_min=0, a_max=linf)


def project_simplex_with_subgroup_l1(x, l1_norm=1, subgroup=None, subgroup_l1=np.Infinity):
    """
    Projection of x onto set where z_i >=0, sum_i z_i <= l1_norm, and sum_subgroup z_i <= subgroup_l1
    :param x: np.ndarray
    :param l1_norm: positive real
    Constraint on l1 norm of x.
    :param subgroup: np.ndarray of indices
    Indices for a subset of elements of x.
    :param subgroup_l1: positive real
    Constraint on l1 norm of a subset of elements of x.
    :return: np.ndarray
    """

    if subgroup is None:
        return project_positive_simplex(x, l1_norm=l1_norm)

    xplus = np.clip(x, a_min=0, a_max=np.Infinity)

    a = np.sum(xplus)
    b = np.sum(xplus[subgroup])

    # Case 1: Both constraints satisfied.
    if a <= l1_norm and b <= subgroup_l1:
        return xplus

    # Case 2: Constraint on subgroup is tight but constraint on x is loose.
    if (subgroup_l1 + (a - b)) <= l1_norm:  # a-b is sum of elements not in the subgroup
        xplus[subgroup] = project_positive_simplex(x[subgroup], l1_norm=subgroup_l1)
        return xplus


    # Case 3: Constraint on x is tight but subgroup constraint is loose.
    xhat = project_positive_simplex(x, l1_norm=l1_norm)
    if np.sum(xhat[subgroup]) <= subgroup_l1:
        return xhat

    not_subgroup = np.delete(np.arange(0, len(x), 1), subgroup)
    # Case 4: Neither constraint satisfied. At solution both constraints are tight.
    xplus[subgroup] = project_positive_simplex(x[subgroup], l1_norm=subgroup_l1)
    xplus[not_subgroup] = project_positive_simplex(x[not_subgroup], l1_norm=l1_norm - subgroup_l1)
    return xplus


def project_positive_simplex(x, l1_norm=1):

    assert type(x) is np.ndarray, "Input should be numeric numpy array."

    xplus = np.clip(x, a_min=0, a_max=np.Infinity)
    if np.sum(xplus) <= l1_norm:
        return xplus

    ind = np.argsort(x)
    xsort = x[ind]

    n = len(x)
    i = n

    tail_sum = 0
    while i > 0:
        i = i - 1
        tail_sum += xsort[i]
        mu = (tail_sum - l1_norm) / (n - i)
        if (i == 0) or (mu >= xsort[i - 1]):
            break

    return np.clip(x - mu, a_min=0, a_max=np.Infinity)


class MetricLearner:
    
    def __init__(self,  stepsize=1.0, solver='dual average', null_if_diverge=True, param_update_frac=1,
                 batch_size=None, penalty='auto', rank_transform=False,
                 l1_constraint=1.0, linf_constraint=1.0,
                 subgroup=None, subgroup_l1=np.Infinity,
                 subgroup_lower=0, subgroup_upper=1,
                 f=lambda u: np.exp(-u), g=lambda u: np.exp(-u), diff=np.abs):
        """
        :param stepsize: Step size used for stochastic gradient descent. In each step we move
        stepsize/sqrt(iter) opposite the direction of the gradient. Defaults to 1.
        :param batch_size: Batch size used for stochastic gradient descent.
        :param penalty: Choices are 'auto', 'none', or 'user_defined'.
        :param rank_transform: Boolean parameter for whether to rank transform the input data.
        :param l1_constraint: Maximum allowable L1 norm of coefficient vector. Defaults to 1.
        :param linf_constraint: Maximum allowable value for a single coefficient. Defaults to Infinity.
        :param subgroup: np.darray of indices
        Indices of a subset of the coefficient vector on which to place an additional l1 norm constraint
        :param subgroup_l1: l1 norm constraint on coefficients of subgroup
        Can be set to None if no L1 constraint.
        :param param_update_frac: Real value between 0 and 1.
        Fraction of parameters to update in each step of optimization (defaults to 1).
        """

        self.solver = solver

        self.rank_transform = rank_transform

        assert f is not None, "Must specify similarity/dissimilarity function, f."
        self.f = f
        assert g is not None, "Must specify g = f' for gradient computations."
        self.g = g

        self.diff = np.abs if diff is None else diff

        self.l1_constraint = l1_constraint
        self.linf_constraint = linf_constraint
        self.subgroup = subgroup
        self.subgroup_lower = subgroup_lower
        self.subgroup_upper = subgroup_upper
        self.subgroup_l1 = subgroup_l1

        assert penalty in set(('auto', 'none', 'user_defined')), "penalty must be 'auto', 'none', or 'user_defined'."
        self.penalty = penalty
        self.lam = None
        self.beta_tol = None # coefficients smaller than this threshold are considered 0

        self.iters = 1
        self.batch_size = 100 if batch_size is None else batch_size
        self.stepsize = stepsize
        self.not_converged = False
        self.null_if_diverge = null_if_diverge
        self.param_update_frac = param_update_frac
        self.active_set_stable = 0  # counter for how many rounds the active set has been stable

        self.n = None
        self.p = None
        self.X = None

        # private variables for keeping track of observation weights and sampling observations according to weight
        self.sample_weight = None  # weight given to each sample point
        self.class_p = None  # conditional probability that a particular observation belongs to class 1
        self.sum_class0 = None  # total sample weight of class 0
        self.sum_class1 = None  # total sample weight of class 1
        self.pairs_frac0 = None  # probability of drawing a 0,0 pair given a draw of concordant observations
        self.p0 = None  # prob of drawing an observation given that a class 0 observation is drawn
        self.p1 = None  # prob of drawing an observation given that a class 1 observation is drawn
        self.sampler0 = None  # RandomSampler for drawing from class 0 according to distribution p0
        self.sampler1 = None  # RandomSampler for drawing from class 1 according to distribution p1

        # kernel weights and dual variables used in dual averaging
        self.beta = None
        self.dual_var = None
        self.old_beta = None
        self.active = None


    def set_X(self, X):
        if self.rank_transform:
            self.X = np.apply_along_axis(rankdata, 0, X) / (self.n + 1)
        else:
            self.X = X

    def set_penalty(self, lam):
        if self.penalty is 'auto':
            self.lam = np.sqrt(np.log(self.p)/self.n)
        elif self.penalty is 'none':
            self.lam = 0
        elif self.penalty is 'user_defined':
            assert isinstance(lam, numbers.Number), "penalty parameter must be a real number."
            self.lam = lam
        else:
            raise ValueError("Unrecognized value for parameter 'penalty'.")

    def init_dual(self, init_value):
        if type(init_value) == np.ndarray:
            self.dual_var = init_value.reshape((self.p,))
        elif isinstance(init_value, numbers.Number):
            self.dual_var = np.ones(self.p)*init_value
        else:
            raise ValueError("Unrecognized initial value for dual variable in dual averaging.")

    def init_beta(self, init_option, init_frac_of_l1=0.8):
        """
        Called by train to initialize coefficient vector (kernel weights).

        :param init_option: Can be either a user supplied vector of same length as number of features or
        one of the options {zero, random, equal}. For random and equal, the sum of the initial weights is
        set to init_frac_of_l1 * l1_constraint.
        :param init_frac_of_l1: Real value between 0 and 1. Initial coefficient sum is set to this fraction of
        the l1 constraint.
        """
        if type(init_option) == np.ndarray:
            self.beta = init_option.reshape((self.p,))
            return

        assert init_option in set(('random', 'equal', 'zero')), "Options for initialization are random, equal, or zero."

        if init_option == 'zero':
            self.beta = np.zeros(self.p)
        elif init_option == 'random':
            self.beta = np.min([self.l1_constraint,1])*np.random.dirichlet(np.ones(self.p)*0.25, size=1).squeeze()*init_frac_of_l1
        elif init_option == 'equal':
            self.beta = np.min([self.l1_constraint,1])*np.ones(self.p)/self.p*init_frac_of_l1
        else:
            raise ValueError("Unrecognized option for initializing beta.")
          
    def compute_g(self, D, beta, active):
        # only compute g using features with coefficient > tol
        if len(active) == 0:
            return np.ones(D.shape[0])*self.g(0)
        else:
            return self.g(np.matmul(D[:, active], beta[active]))

    def _compute_f(self, D):
        active = self.beta > self.beta_tol
        return self.f(np.matmul(D[:, active], self.beta[active]))

    def eval_objective(self):

        d_discord = self.sample_diff(self.X, self.X, self.sampler0, self.sampler1)
        d_concord0 = self.sample_diff(self.X, self.X, self.sampler0, self.sampler0)
        d_concord1 = self.sample_diff(self.X, self.X, self.sampler1, self.sampler1)

        f_discord = np.average(self._compute_f(d_discord))
        f_concord0 = np.average(self._compute_f(d_concord0))
        f_concord1 = np.average(self._compute_f(d_concord1))

        return self.pairs_frac0*f_concord0 + (1-self.pairs_frac0)*f_concord1 - f_discord

    def _is_converged(self, old, new, convergence_tol=1e-4):
        #chg = np.sum(np.abs(old - new))/(np.sum(np.abs(old)) + 1e-10)

        old_active = np.where(old >= self.beta_tol)[0]
        new_active = np.where(new >= self.beta_tol)[0]

        overlap = len(np.intersect1d(old_active, new_active))/np.max([len(old_active), len(new_active), 1])
        #if chg < convergence_tol:

        if overlap == 1:
            self.active_set_stable += 1

        if self.active_set_stable > 2:
            return True
        else:
            return False

    def sample_pairs(self, sampler0, sampler1):
        """
        Sample pairs (i,j) with replacement with probability proportional to w0_i*w0_j
        :param num_pairs: Integer. Number of pairs to sample.
        :param sampler0: RandomSampler
        :param sampler1: RandomSampler
        :return: pair1, pair2. Two 1-D arrays containing indices for sampled pairs.
        """

        pair1 = sampler0.next()
        pair2 = sampler1.next()

        return pair1, pair2

    def sample_diff(self, arr0, arr1, sampler0, sampler1):
        """
        Randomly sample squared differences diff(arr0[i,:] - arr1[j,:]) from 2-D arrays arr0 and arr1
        according to row weights specified by RandomSamplers sampler0 and sampler1. The diff function
        is specified when initializing a MetricLearner instance and defaults to absolute value function.
        :param arr0: 2-D array
        :param arr1: 2-D array
        :param sampler0: RandomSampler
        :param sampler1: RandomSampler
        """

        n0, p0 = arr0.shape
        n1, p1 = arr1.shape

        assert p0 == p1, "A and B must same column dimension."

        pair0, pair1 = self.sample_pairs(sampler0, sampler1)
        return self.diff(arr0[pair0, :] - arr1[pair1, :])

    def get_stochastic_grad(self, X, beta, active, sampler0, sampler1, lam):

        # between group difference
        d_discord = self.sample_diff(arr0=X, arr1=X, sampler0=sampler0, sampler1=sampler1)
        g_discord = self.compute_g(D=d_discord, beta=beta, active=active).reshape((-1,1))
        grad_discord = np.average(d_discord*g_discord, axis=0)  # d_discord * g_discord uses numpy broadcasting

        # within group difference for class 0
        d_concord0 = self.sample_diff(arr0=X, arr1=X, sampler0=sampler0, sampler1=sampler0)
        g_concord0 = self.compute_g(D=d_concord0, beta=beta, active=active).reshape((-1,1))
        grad_concord0 = np.average(d_concord0*g_concord0, axis=0)

        # within group difference for class 1
        d_concord1 = self.sample_diff(arr0=X, arr1=X, sampler0=sampler1, sampler1=sampler1)
        g_concord1 = self.compute_g(D=d_concord1, beta=beta, active=active).reshape((-1,1))
        grad_concord1 = np.average(d_concord1*g_concord1, axis=0)

        return self.pairs_frac0*grad_concord0 + (1-self.pairs_frac0)*grad_concord1 - grad_discord + lam

    def get_vars(self):
        #if self.null_if_diverge and self.not_converged:
        #    return np.array([], dtype=np.int)  # if not converged, return no variables for safety
        #else:
        return np.where(self.beta >= self.beta_tol)[0]

    def init_sample_weight(self, class_p, sample_weight=None):
        """
        Each observation can be weighted differently. By default, balanced class weighting.
        :param class_p: numpy array with class proportions
        :param sample_weight: Numpy array with same length as number of observations.
        All entries must be positive. Entries not required to sum to 1.
        """
        if sample_weight is None:
            p1 = np.sum(class_p)/self.n
            p0 = np.sum(1-class_p)/self.n

            self.sample_weight = class_p * p0 + (1-class_p) * p1
            self.sample_weight = self.sample_weight / np.sum(self.sample_weight)
        else:
            assert len(sample_weight) == self.n, "Length of sample weight must be same as number of observations."
            self.sample_weight = sample_weight

    def init_response_weight(self, class_p):
        """
        Responses are allowed to be "soft" i.e. probabilities P(Y=1|x) rather than binary (0-1)
        values. The probability that the ith observation is sampled and its response is 1 is
        proportional to

        sample_weight[i] * class_p[i].

        Similarly the probability that ith observation is sampled and its response is 0 is proportional to

        sample_weight[i] * (1-class_p[i]).

        These two quantities can be normalized to give the probability of sampling an observation
        conditional on being in class 1 or class 0 respectively.
        :param class_p: Numpy array with each entry between 0 and 1. Each try represents P(Y=1|x).

        """
        assert 0 <= np.min(class_p) and np.max(class_p) <= 1, "Probabilities must be between 0 and 1."
        assert self.sample_weight is not None, "Initialize sample weights first before response weight."

        # class_p contains probability that observation i has response equal to 1
        self.class_p = class_p
        self.sum_class1 = np.sum(self.class_p * self.sample_weight)  # total weight of class 1
        self.sum_class0 = np.sum((1 - self.class_p) * self.sample_weight)  # total weight of class 0

        # probability of drawing a within class pair in which both responses are 0 (other
        # possibility is that both responses are 1)
        self.pairs_frac0 = self.sum_class0 ** 2 / (self.sum_class0 ** 2 + self.sum_class1 ** 2)

        # probability of drawing observation given it is in class 0
        self.p0 = (1 - self.class_p) * self.sample_weight / self.sum_class0
        self.sampler0 = random_sampler.RandomSampler(self.n, batch_size=self.batch_size, weight=self.p0)

        # probability of drawing observation given it is in class 1
        self.p1 = self.class_p * self.sample_weight / self.sum_class1
        self.sampler1 = random_sampler.RandomSampler(self.n, batch_size=self.batch_size, weight=self.p1)

    def reset(self):
        """
        Call reset after performing screening along a path and before starting new path.
        """
        self.iters = 1
        self.beta = None
        self.sample_weight = None
        self.subgroup = None

    def reset_beta(self):
        self.beta = None

    def train(self, X, class_p, sample_weight=None, subgroup=None, subgroup_l1=None, max_iter=5000,
              beta_init='random', lam=None, verbose=False, zero_tol=1e-6, convergence_tol=1e-4, warm_start=False,
              subgroup_lower=None, subgroup_upper=None, dual_init=0.0, warmup=100):

        self.n, self.p = X.shape
        self.set_X(X)

        self.init_sample_weight(class_p=class_p, sample_weight=sample_weight)
        self.init_response_weight(class_p=class_p)

        self.warmup = warmup  # only check convergence after warmup has been satisfied

        # initialize coefficient vector
        if self.beta is None or not warm_start:
            self.init_beta(beta_init)
        else:
            assert len(self.beta) == self.p, "Warm start set to true but length of coefficient vector not equal to\
            number of features in X matrix."
            self.warmup = 25

        self.beta_tol = np.min([convergence_tol*100, 1e-2])
        self.old_beta = self.beta.copy()
        self.init_dual(init_value=dual_init)

        # initialize active variables
        self.active = np.nonzero(self.beta > 0)[0]

        self.set_penalty(lam)
        self.subgroup = subgroup
        if subgroup_l1 is not None: self.subgroup_l1 = subgroup_l1
        if subgroup_lower is not None: self.subgroup_lower = subgroup_lower
        if subgroup_upper is not None: self.subgroup_upper = subgroup_upper

        self.active_set_stable = 0
        self.iters = 1

        self.continue_training(max_iter=max_iter, verbose=verbose, zero_tol=zero_tol, convergence_tol=convergence_tol)

    def _get_steepest_gradients(self, grad):
        cutoff = np.percentile(grad, q=self.param_update_frac*100)
        grad[(grad > cutoff) & (grad < 0)] = 0
        return grad

    def _dual_average_update(self, dual_var, grad, stepsize, iter_count):
        dual_var -= stepsize/np.sqrt(iter_count) * grad
        beta = project_simplex_with_box_constraints(dual_var, l1_norm=self.l1_constraint, linf=self.linf_constraint)
        if self.subgroup is not None:
            beta[self.subgroup] = np.clip(beta[self.subgroup], a_min=self.subgroup_lower, a_max=self.subgroup_upper)
        return beta

    def _sgd_update(self, grad, stepsize, iter_count):
        self.beta -= stepsize/np.sqrt(iter_count) * grad
        self.beta = project_simplex_with_box_constraints(self.beta, l1_norm=self.l1_constraint, linf=self.linf_constraint)

    def continue_training(self, max_iter, verbose=False, zero_tol=1e-6, convergence_tol=1e-4):

        for i in range(max_iter):
            self.active = np.nonzero(self.beta > 0)[0]
            grad = self.get_stochastic_grad(X=self.X, beta=self.beta, active=self.active,
                                            sampler0=self.sampler0, sampler1=self.sampler1, lam=self.lam)

            if self.param_update_frac < 1:
                grad = self._get_steepest_gradients(grad)  # option to update only steepest gradients

            if self.solver == 'dual average':
                self.beta = self._dual_average_update(dual_var=self.dual_var, grad=grad, stepsize=self.stepsize,
                                                      iter_count=self.iters)
            else:
                self._sgd_update(grad=grad, stepsize=self.stepsize, iter_count=self.iters)

            if (i > self.warmup) and (i % 25 == 0) and ((np.sum(np.abs(self.beta)) < zero_tol) or
                                  self._is_converged(old=self.old_beta, new=self.beta, convergence_tol=convergence_tol)):
                break

            if i % 25 == 0:
                self.old_beta = self.beta.copy()

            self.iters += 1
            if verbose and (i % 100 == 0):
                print(np.round(self.beta, decimals=3))

        print("Training lasted for %d iterations." % self.iters, flush=True)

        if i == (max_iter-1):
            self.not_converged = True
            print('Maximum iteration reached without convergence. Set max_iter higher.', flush=True)


    def hier_train(self, init_value, beta=None, active_set=np.array([], dtype=np.int64), verbose=False, zero_tol=1e-6, \
                   convergence_tol=1e-4, marginal_sampling_size=5000):
        """

        :param active_set: p, boolean array-like
        Boolean array indicating which variables have already been selected at the start of hierarchical training.
        """

        if beta is None: beta = np.zeros(self.p)
        if len(active_set) > 0: beta[active_set] = init_value

        # we use a larger sampling size in the first step to determine whether any gradients are negative
        sampler0 = random_sampler.RandomSampler(self.n, batch_size=marginal_sampling_size, weight=self.p0, buf=3)
        sampler1 = random_sampler.RandomSampler(self.n, batch_size=marginal_sampling_size, weight=self.p1, buf=3)
        grad = self.get_stochastic_grad(X=self.X, beta=beta, active=active_set, sampler0=sampler0, sampler1=sampler1,\
                                        lam=self.lam)
        chosen = np.nonzero(grad < 0)[0]

        # check if any new variables have been added
        chosen = np.union1d(chosen, active_set)
        if len(chosen) == len(active_set):
            return active_set

        return self.hier_train(init_value=init_value, beta=beta, active_set=chosen, verbose=verbose, zero_tol=zero_tol,\
                               convergence_tol=convergence_tol, marginal_sampling_size=marginal_sampling_size)